@@ -5,32 +5,29 @@
 # pylint: disable=invalid-name,protected-access
 """ffpuppet tests"""
 import errno
+from http.server import HTTPServer, BaseHTTPRequestHandler
 import os
 import platform
 import shutil
 import socket
 import stat
-from subprocess import call, check_output
+from subprocess import call, check_output, Popen
 import threading
 import time
-try:  # py 2-3 compatibility
-    from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler  # pylint: disable=import-error
-except ImportError:
-    from http.server import HTTPServer, BaseHTTPRequestHandler  # pylint: disable=import-error
 
 from psutil import AccessDenied, NoSuchProcess, Process, wait_procs
 import pytest
 
+from .bootstrapper import Bootstrapper
 from .core import FFPuppet
-from .exceptions import BrowserTimeoutError, BrowserTerminatedError, LaunchError
+from .exceptions import BrowserTimeoutError, BrowserTerminatedError, LaunchError, TerminateError
 from .helpers import get_processes
 from .minidump_parser import MinidumpParser
 
+Bootstrapper.POLL_WAIT = 0.2
 CWD = os.path.realpath(os.path.dirname(__file__))
 TESTFF_BIN = os.path.join(CWD, "resources", "testff.py")
-TESTMDSW_BIN = os.path.join(CWD, "resources", "testmdsw.py")
 
-MinidumpParser.MDSW_BIN = TESTMDSW_BIN
 MinidumpParser.MDSW_MAX_STACK = 8
 
 
@@ -118,10 +115,10 @@ def test_ffpuppet_02(tmp_path):
         assert ffp.launches == 0
         assert ffp.reason == ffp.RC_ALERT
 
-def test_ffpuppet_03(tmp_path):
+def test_ffpuppet_03(mocker, tmp_path):
     """test hang on start"""
+    mocker.patch("ffpuppet.core.FFPuppet.LAUNCH_TIMEOUT_MIN", 1)
     with FFPuppet() as ffp:
-        ffp.LAUNCH_TIMEOUT_MIN = 1
         prefs = (tmp_path / "prefs.js")
         prefs.write_bytes(b"//fftest_startup_hang\n")
         start = time.time()
@@ -130,7 +127,7 @@ def test_ffpuppet_03(tmp_path):
         duration = time.time() - start
         ffp.close()
         assert ffp.reason == ffp.RC_CLOSED
-        assert duration >= ffp.LAUNCH_TIMEOUT_MIN
+        assert duration >= 1
         assert duration < 30
 
 def test_ffpuppet_04(tmp_path):
@@ -271,8 +268,6 @@ def test_ffpuppet_11(tmp_path):
     with FFPuppet() as ffp:
         ffp.add_abort_token(r"TEST\dREGEX\.+")
         ffp.add_abort_token("simple_string")
-        with pytest.raises(AssertionError):
-            ffp.add_abort_token(None)
         ffp.add_abort_token(r"ASSERTION:\s\w+")
         with HTTPTestServer() as srv:
             ffp.launch(TESTFF_BIN, location=srv.get_addr(), prefs_js=str(prefs))
@@ -349,20 +344,11 @@ def test_ffpuppet_15(tmp_path):
         assert not location.startswith("/")
         assert os.path.normpath(os.path.join("/", location)) == fname
 
-def test_ffpuppet_16():
-    """test passing nonexistent file to launch() via prefs_js"""
-    with FFPuppet() as ffp:
-        with pytest.raises(IOError, match="prefs.js file does not exist"):
-            ffp.launch(TESTFF_BIN, prefs_js="missing.js")
 
-@pytest.mark.skipif(platform.system() == "Linux" and call(["which", "gdb"]),
+@pytest.mark.skipif(platform.system() != "Linux" or call(["which", "gdb"]),
                     reason="GDB not installed")
-def test_ffpuppet_17(tmp_path):
+def test_ffpuppet_16(tmp_path):
     """test launching with gdb"""
-    if platform.system() != "Linux":
-        with pytest.raises(EnvironmentError, match="GDB is only supported on Linux"):
-            FFPuppet(use_gdb=True)
-        return
     with FFPuppet(use_gdb=True) as ffp:
         bin_path = str(check_output(["which", "echo"]).strip().decode("ascii"))
         # launch will fail b/c 'echo' will exit right away but that's fine
@@ -376,7 +362,7 @@ def test_ffpuppet_17(tmp_path):
         assert b"[Inferior " in log_data
         assert b"quit_with_code" in log_data
 
-def test_ffpuppet_18(tmp_path):
+def test_ffpuppet_17(tmp_path):
     """test calling save_logs() before close()"""
     with FFPuppet() as ffp:
         with HTTPTestServer() as srv:
@@ -384,19 +370,12 @@ def test_ffpuppet_18(tmp_path):
             with pytest.raises(AssertionError):
                 ffp.save_logs(str(tmp_path / "logs"))
 
-@pytest.mark.skipif(platform.system() == "Linux" and call(["which", "valgrind"]),
+@pytest.mark.skipif(platform.system() != "Linux" or call(["which", "valgrind"]),
                     reason="Valgrind not installed")
-def test_ffpuppet_19(tmp_path):
+def test_ffpuppet_18(tmp_path):
     """test launching with Valgrind"""
-    if platform.system() != "Linux":
-        with pytest.raises(EnvironmentError, match="Valgrind is only supported on Linux"):
-            FFPuppet(use_valgrind=True)
-        return
     vmv = FFPuppet.VALGRIND_MIN_VERSION
     try:
-        FFPuppet.VALGRIND_MIN_VERSION = 9999999999.99
-        with pytest.raises(EnvironmentError, match=r"Valgrind >= \d+\.\d+ is required"):
-            FFPuppet(use_valgrind=True)
         FFPuppet.VALGRIND_MIN_VERSION = 0
         with FFPuppet(use_valgrind=True) as ffp:
             bin_path = str(check_output(["which", "echo"]).strip().decode("ascii"))
@@ -412,7 +391,7 @@ def test_ffpuppet_19(tmp_path):
     finally:
         FFPuppet.VALGRIND_MIN_VERSION = vmv
 
-def test_ffpuppet_20(tmp_path):
+def test_ffpuppet_19(tmp_path):
     """test detecting invalid prefs file"""
     prefs = (tmp_path / "prefs.js")
     prefs.write_bytes(b"//fftest_invalid_js\n")
@@ -421,7 +400,7 @@ def test_ffpuppet_20(tmp_path):
             with pytest.raises(LaunchError, match="'.+?' is invalid"):
                 ffp.launch(TESTFF_BIN, location=srv.get_addr(), prefs_js=str(prefs))
 
-def test_ffpuppet_21():
+def test_ffpuppet_20():
     """test log_length()"""
     with FFPuppet() as ffp:
         assert ffp.log_length("INVALID") is None
@@ -434,7 +413,7 @@ def test_ffpuppet_21():
         # verify clean_up() removed the logs
         assert ffp.log_length("stderr") is None
 
-def test_ffpuppet_22():
+def test_ffpuppet_21():
     """test running multiple instances in parallel"""
     ffps = list()
     try:
@@ -453,7 +432,7 @@ def test_ffpuppet_22():
         for ffp in ffps:
             ffp.clean_up()
 
-def test_ffpuppet_23(tmp_path):
+def test_ffpuppet_22(tmp_path):
     """test hitting log size limit"""
     prefs = (tmp_path / "prefs.js")
     prefs.write_bytes(b"//fftest_big_log\n")
@@ -473,7 +452,7 @@ def test_ffpuppet_23(tmp_path):
         assert sum(x.stat().st_size for x in logfiles) > limit
         assert b"LOG_SIZE_LIMIT_EXCEEDED" in (logs / "log_ffp_worker_log_size.txt").read_bytes()
 
-def test_ffpuppet_24(tmp_path):
+def test_ffpuppet_23(tmp_path):
     """test collecting and cleaning up ASan logs"""
     test_logs = list()
     with FFPuppet() as ffp:
@@ -514,18 +493,21 @@ def test_ffpuppet_24(tmp_path):
                 assert log_fp.readline() in ("BAD LOG\n", "GOOD LOG\n", "SHORT LOG\n")
     assert not any(os.path.isfile(f) for f in test_logs)
 
-def test_ffpuppet_25(tmp_path):
+def test_ffpuppet_24(mocker, tmp_path):
     """test multiple minidumps"""
+    def _fake_process_minidumps(dmps, _, add_log):
+        for num, _ in enumerate(x for x in os.listdir(dmps) if x.endswith(".dmp")):
+            lfp = add_log("minidump_%02d" % (num + 1,))
+            lfp.write(b"test")
+    mocker.patch("ffpuppet.core.process_minidumps", side_effect=_fake_process_minidumps)
     profile = (tmp_path / "profile")
     profile.mkdir()
     (profile / "minidumps").mkdir()
-    # 'symbols' directory needs to exist to satisfy a check
-    (profile / "symbols").mkdir()
     with FFPuppet(use_profile=str(profile)) as ffp:
         ffp.launch(TESTFF_BIN)
-        ffp._last_bin_path = ffp.profile
+        ffp._bin_path = ffp.profile
         # create "test.dmp" files
-        md_path = os.path.join(ffp._last_bin_path, "minidumps")
+        md_path = os.path.join(ffp._bin_path, "minidumps")
         with open(os.path.join(md_path, "test1.dmp"), "w") as out_fp:
             out_fp.write("1a\n1b")
         with open(os.path.join(md_path, "test2.dmp"), "w") as out_fp:
@@ -540,7 +522,7 @@ def test_ffpuppet_25(tmp_path):
         assert any(logs.glob("log_minidump_02.txt"))
         assert any(logs.glob("log_minidump_03.txt"))
 
-def test_ffpuppet_26(tmp_path):
+def test_ffpuppet_25(tmp_path):
     """test multiprocess target"""
     prefs = (tmp_path / "prefs.js")
     prefs.write_bytes(b"//fftest_multi_proc\n")
@@ -558,7 +540,7 @@ def test_ffpuppet_26(tmp_path):
         assert not ffp.is_running()
         assert ffp.wait(timeout=0)
 
-def test_ffpuppet_27(tmp_path):
+def test_ffpuppet_26(tmp_path):
     """test multiprocess (target terminated)"""
     prefs = (tmp_path / "prefs.js")
     prefs.write_bytes(b"//fftest_multi_proc\n")
@@ -580,14 +562,10 @@ def test_ffpuppet_27(tmp_path):
         assert not ffp.is_running()
         assert ffp.wait(timeout=0)
 
-@pytest.mark.skipif(platform.system() == "Linux" and call(["which", "rr"]),
+@pytest.mark.skipif(platform.system() != "Linux" or call(["which", "rr"]),
                     reason="rr not installed")
-def test_ffpuppet_28(tmp_path):
+def test_ffpuppet_27(tmp_path):
     """test launching with rr"""
-    if platform.system() != "Linux":
-        with pytest.raises(EnvironmentError, match="rr is only supported on Linux"):
-            FFPuppet(use_rr=True)
-        return
     # NOTE: this can hang if ptrace is blocked by seccomp
     if call(["rr", "record", "echo"]) != 0:
         pytest.skip("Environment not configured to run rr")
@@ -605,7 +583,7 @@ def test_ffpuppet_28(tmp_path):
         assert b"rr record" in log_data
         assert b"[ffpuppet] Reason code:" in log_data
 
-def test_ffpuppet_29(tmp_path):
+def test_ffpuppet_28(tmp_path):
     """test rmtree error handler"""
     # normal profile creation
     # - just create a puppet, write a readonly file in its profile, then call close()
@@ -631,7 +609,7 @@ def test_ffpuppet_29(tmp_path):
         ffp.close()
         assert not os.path.isdir(prof_path)
 
-def test_ffpuppet_30(tmp_path):
+def test_ffpuppet_29(tmp_path):
     """test using a readonly prefs.js and extension"""
     prefs = (tmp_path / "prefs.js")
     prefs.touch()
@@ -645,12 +623,12 @@ def test_ffpuppet_30(tmp_path):
         ffp.close()
         assert not os.path.isdir(prof_path)
 
-def test_ffpuppet_31(tmp_path):
+def test_ffpuppet_30(mocker, tmp_path):
     """test _crashreports()"""
     class StubbedLaunch(FFPuppet):
         def __init__(self):
-            super(StubbedLaunch, self).__init__()
-            self._use_valgrind = True
+            super().__init__()
+            self._dbg = FFPuppet.DBG_VALGRIND
         def launch(self):  # pylint: disable=arguments-differ
             profile = (tmp_path / "profile")
             profile.mkdir()
@@ -663,29 +641,43 @@ def test_ffpuppet_31(tmp_path):
     with StubbedLaunch() as ffp:
         ffp.launch()
         assert not any(ffp._crashreports())
+        # ignored sanitizer warnings
         ign_log = "%s.1" % (ffp._logs.PREFIX_SAN,)
-        san_log = "%s.2" % (ffp._logs.PREFIX_SAN,)
-        vg1_log = "%s.1" % (ffp._logs.PREFIX_VALGRIND,)
-        vg2_log = "%s.2" % (ffp._logs.PREFIX_VALGRIND,)
         with open(os.path.join(ffp._logs.working_path, ign_log), "w") as ofp:
-            ofp.write("==123==WARNING: Symbolizer buffer too small\n")
             ofp.write("==123==WARNING: Symbolizer buffer too small\n\n")
+            ofp.write("==123==WARNING: Symbolizer buffer too small\n\n")
+        # valid sanitizer log
+        san_log = "%s.2" % (ffp._logs.PREFIX_SAN,)
         with open(os.path.join(ffp._logs.working_path, san_log), "w") as ofp:
             ofp.write("test\n")
+        # valid Valgrind log - with error
+        vg1_log = "%s.1" % (ffp._logs.PREFIX_VALGRIND,)
         with open(os.path.join(ffp._logs.working_path, vg1_log), "w") as ofp:
             ofp.write("test\n")
+        # valid Valgrind log - without error
+        vg2_log = "%s.2" % (ffp._logs.PREFIX_VALGRIND,)
         with open(os.path.join(ffp._logs.working_path, vg2_log), "w") as ofp:
             pass
+        # nothing interesting
         with open(os.path.join(ffp._logs.working_path, "junk.log"), "w") as ofp:
             ofp.write("test\n")
+        # valid minidump
         with open(os.path.join(ffp.profile, "minidumps", "test.dmp"), "w") as ofp:
             ofp.write("test\n")
+        # nothing interesting
         with open(os.path.join(ffp.profile, "minidumps", "test.junk"), "w") as ofp:
             pass
+        assert not ffp._logs.watching
         assert len(list(ffp._crashreports())) == 3
+        assert ffp._logs.watching
         assert len(list(ffp._crashreports(skip_md=True))) == 2
+        # fail to open and scan sanitizer file
+        ffp._logs.watching.clear()
+        mocker.patch("ffpuppet.core.open", side_effect=OSError)
+        assert len(list(ffp._crashreports())) == 4
+        assert not ffp._logs.watching
 
-def test_ffpuppet_32(tmp_path):
+def test_ffpuppet_31(tmp_path):
     """test build_launch_cmd()"""
     with FFPuppet() as ffp:
         cmd = ffp.build_launch_cmd("bin_path", ["test"])
@@ -693,19 +685,17 @@ def test_ffpuppet_32(tmp_path):
         assert cmd[0] == "bin_path"
         assert cmd[-1] == "test"
         # GDB
-        ffp._use_gdb = True
+        ffp._dbg = FFPuppet.DBG_GDB
         cmd = ffp.build_launch_cmd("bin_path")
         assert len(cmd) > 2
         assert cmd[0] == "gdb"
-        ffp._use_gdb = False
         # RR
-        ffp._use_rr = True
+        ffp._dbg = FFPuppet.DBG_RR
         cmd = ffp.build_launch_cmd("bin_path")
         assert len(cmd) > 2
         assert cmd[0] == "rr"
-        ffp._use_rr = False
         # Valgrind
-        ffp._use_valgrind = True
+        ffp._dbg = FFPuppet.DBG_VALGRIND
         try:
             os.environ["VALGRIND_SUP_PATH"] = "blah"
             with pytest.raises(IOError):
@@ -718,9 +708,8 @@ def test_ffpuppet_32(tmp_path):
             assert cmd[0] == "valgrind"
         finally:
             os.environ.pop("VALGRIND_SUP_PATH")
-        ffp._use_valgrind = False
 
-def test_ffpuppet_33():
+def test_ffpuppet_32():
     """test cpu_usage()"""
     with FFPuppet() as ffp:
         assert not any(ffp.cpu_usage())
@@ -734,3 +723,164 @@ def test_ffpuppet_33():
             assert usage[0][1] >= 0
         ffp.close()
         assert ffp.wait(timeout=10)
+
+def test_ffpuppet_33(mocker):
+    """test _dbg_sanity_check()"""
+    fake_system = mocker.patch("ffpuppet.core.system", autospec=True)
+    fake_chkout = mocker.patch("ffpuppet.core.check_output", autospec=True)
+    # gdb - success
+    fake_system.return_value = "Linux"
+    FFPuppet._dbg_sanity_check(FFPuppet.DBG_GDB)
+    assert fake_chkout.call_count == 1
+    fake_chkout.reset_mock()
+    # gdb - not installed
+    fake_chkout.side_effect = OSError
+    with pytest.raises(EnvironmentError, match="Please install GDB"):
+        FFPuppet._dbg_sanity_check(FFPuppet.DBG_GDB)
+    fake_chkout.reset_mock()
+    fake_chkout.side_effect = None
+    # gdb - unsupported OS
+    fake_system.return_value = "Windows"
+    with pytest.raises(EnvironmentError, match="GDB is only supported on Linux"):
+        FFPuppet._dbg_sanity_check(FFPuppet.DBG_GDB)
+    # rr - success
+    fake_system.return_value = "Linux"
+    FFPuppet._dbg_sanity_check(FFPuppet.DBG_RR)
+    assert fake_chkout.call_count == 1
+    fake_chkout.reset_mock()
+    # rr - not installed
+    fake_chkout.side_effect = OSError
+    with pytest.raises(EnvironmentError, match="Please install rr"):
+        FFPuppet._dbg_sanity_check(FFPuppet.DBG_RR)
+    fake_chkout.reset_mock()
+    fake_chkout.side_effect = None
+    # rr - unsupported OS
+    fake_system.return_value = "Windows"
+    with pytest.raises(EnvironmentError, match="rr is only supported on Linux"):
+        FFPuppet._dbg_sanity_check(FFPuppet.DBG_RR)
+    # valgrind - success
+    fake_system.return_value = "Linux"
+    fake_chkout.return_value = b"valgrind-%0.2f" % (FFPuppet.VALGRIND_MIN_VERSION)
+    FFPuppet._dbg_sanity_check(FFPuppet.DBG_VALGRIND)
+    assert fake_chkout.call_count == 1
+    fake_chkout.reset_mock()
+    # valgrind - old version
+    fake_system.return_value = "Linux"
+    fake_chkout.return_value = b"valgrind-0.1"
+    with pytest.raises(EnvironmentError, match=r"Valgrind >= \d+\.\d+ is required"):
+        FFPuppet._dbg_sanity_check(FFPuppet.DBG_VALGRIND)
+    assert fake_chkout.call_count == 1
+    fake_chkout.reset_mock()
+    # valgrind - not installed
+    fake_chkout.side_effect = OSError
+    with pytest.raises(EnvironmentError, match="Please install Valgrind"):
+        FFPuppet._dbg_sanity_check(FFPuppet.DBG_VALGRIND)
+    fake_chkout.reset_mock()
+    fake_chkout.side_effect = None
+    # valgrind - unsupported OS
+    fake_system.return_value = "Windows"
+    with pytest.raises(EnvironmentError, match="Valgrind is only supported on Linux"):
+        FFPuppet._dbg_sanity_check(FFPuppet.DBG_VALGRIND)
+
+def test_ffpuppet_34(mocker):
+    """test _terminate()"""
+    procs = [
+        mocker.Mock(spec=Process, pid=123),
+        mocker.Mock(spec=Process, pid=124)
+    ]
+    mocker.patch("ffpuppet.core.get_processes", autospec=True, return_value=procs)
+    fake_wait_procs = mocker.patch("ffpuppet.core.wait_procs", autospec=True)
+    # successful call to terminate
+    fake_wait_procs.return_value = ([], [])
+    FFPuppet._terminate(1234)
+    assert sum(x.terminate.call_count for x in procs) == 2
+    assert not sum(x.kill.call_count for x in procs)
+    for proc in procs:
+        proc.reset_mock()
+    # successful call to kill
+    fake_wait_procs.return_value = None
+    fake_wait_procs.side_effect = (([], procs), ([], []))
+    FFPuppet._terminate(1234)
+    assert sum(x.terminate.call_count for x in procs) == 2
+    assert sum(x.kill.call_count for x in procs) == 2
+    for proc in procs:
+        proc.reset_mock()
+    # successful call to kill
+    fake_wait_procs.return_value = ([], procs)
+    fake_wait_procs.side_effect = None
+    with pytest.raises(TerminateError):
+        FFPuppet._terminate(1234)
+
+def test_ffpuppet_35(mocker, tmp_path):
+    """test FFPuppet.close() setting reason"""
+    class StubbedProc(FFPuppet):
+        def __init__(self):
+            super().__init__()
+        def launch(self):  # pylint: disable=arguments-differ
+            self.reason = None
+            self._bin_path = str(tmp_path)
+            self._logs.reset()
+            self._proc = mocker.Mock(spec=Popen, pid=123)
+            self._proc.poll.return_value = None
+            profile = (tmp_path / "profile")
+            profile.mkdir(exist_ok=True)
+            self.profile = str(profile)
+        @staticmethod
+        def _terminate(pid, _=None):
+            assert isinstance(pid, int)
+    fake_reports = mocker.patch("ffpuppet.core.FFPuppet._crashreports", autospec=True)
+    fake_reports.return_value = ()
+    fake_wait_files = mocker.patch("ffpuppet.core.wait_on_files", autospec=True)
+    fake_wait_files.return_value = True
+    # process exited - no crash
+    with StubbedProc() as ffp:
+        ffp.launch()
+        ffp._proc.poll.return_value = 0
+        ffp.close()
+        assert ffp._proc is None
+        assert ffp._logs.closed
+        assert ffp.reason == FFPuppet.RC_EXITED
+    # process exited - exit code - crash
+    with StubbedProc() as ffp:
+        ffp.launch()
+        ffp._proc.poll.return_value = -11
+        ffp.close()
+        assert ffp._proc is None
+        assert ffp._logs.closed
+        assert ffp.reason == FFPuppet.RC_ALERT
+    # process running - no crash reports
+    with StubbedProc() as ffp:
+        ffp.launch()
+        ffp._proc.poll.return_value = None
+        ffp.close()
+        assert ffp._proc is None
+        assert ffp._logs.closed
+        assert ffp.reason == FFPuppet.RC_CLOSED
+    # process running - with crash reports, hang waiting to close
+    (tmp_path / "fake_report1").touch()
+    with StubbedProc() as ffp:
+        ffp.launch()
+        ffp._proc.poll.return_value = None
+        fake_reports.return_value = (str(tmp_path / "fake_report1"),)
+        fake_wait_files.return_value = False
+        ffp.close()
+        assert ffp._proc is None
+        assert ffp._logs.closed
+        assert ffp.reason == FFPuppet.RC_ALERT
+    # process running - with crash reports, multiple logs
+    (tmp_path / "fake_report2").touch()
+    with StubbedProc() as ffp:
+        ffp.launch()
+        ffp._proc.poll.return_value = None
+        fake_reports.return_value = None
+        fake_reports.side_effect = (
+            (str(tmp_path / "fake_report1"),),
+            (str(tmp_path / "fake_report1"), str(tmp_path / "fake_report2"),),
+            (str(tmp_path / "fake_report1"), str(tmp_path / "fake_report2"),),
+            (str(tmp_path / "fake_report1"), str(tmp_path / "fake_report2"),),
+            (str(tmp_path / "fake_report1"), str(tmp_path / "fake_report2"),))
+        fake_wait_files.return_value = True
+        ffp.close()
+        assert ffp._proc is None
+        assert ffp._logs.closed
+        assert ffp.reason == FFPuppet.RC_ALERT
